# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the ethereum module."""

import copy
import hashlib
import logging
import math
import random
import re
import tempfile
import time
from pathlib import Path
from typing import Dict, Generator, Optional, Tuple, Union, cast
from unittest import mock
from unittest.mock import MagicMock, Mock, patch

import pytest
from aea_ledger_ethereum import (
    AttributeDictTranslator,
    EthereumApi,
    EthereumCrypto,
    EthereumFaucetApi,
    EthereumHelper,
    SimpleCacheLockWrapper,
    get_gas_price_strategy,
    get_gas_price_strategy_eip1559,
    requests,
    rpc_gas_price_strategy_wrapper,
)
from aea_ledger_ethereum.ethereum import (
    DEFAULT_EIP1559_STRATEGY,
    DEFAULT_GAS_STATION_STRATEGY,
    EIP1559,
    EIP1559_POLYGON,
    GAS_STATION,
    TIP_INCREASE,
    estimate_priority_fee,
    get_base_fee_multiplier,
    get_gas_price_strategy_eip1559_polygon,
)
from requests import HTTPError
from web3 import Web3
from web3._utils.request import _session_cache as session_cache
from web3.datastructures import AttributeDict
from web3.exceptions import ContractLogicError

from aea.common import JSONLike
from aea.crypto.helpers import DecryptError, KeyIsIncorrect

from tests.conftest import DEFAULT_GANACHE_CHAIN_ID, MAX_FLAKY_RERUNS, ROOT_DIR


def get_default_gas_strategies() -> Dict:
    """Returns default gas price strategy."""
    return {
        "default_gas_price_strategy": "eip1559",
        "gas_price_strategies": {
            "gas_station": DEFAULT_GAS_STATION_STRATEGY,
            "eip1559": DEFAULT_EIP1559_STRATEGY,
        },
    }


def get_history_data(n_blocks: int, base_multiplier: int = 100) -> Dict:
    """Returns dummy blockchain history data."""

    return {
        "oldestBlock": 1,
        "reward": [
            [math.ceil(random.random() * base_multiplier) * 1e1]
            for _ in range(n_blocks)
        ],
        "baseFeePerGas": [
            math.ceil(random.random() * base_multiplier) * 1e9 for _ in range(n_blocks)
        ],
    }


def test_attribute_dict_translator():
    """Test the AttributeDictTranslator."""
    di = {
        "1": None,
        "2": True,
        "3": b"some",
        "4": 0.1,
        "5": [1, None, True, {}],
        "6": {"hex": "0x01"},
    }
    res = AttributeDictTranslator.from_dict(di)
    assert AttributeDictTranslator.to_dict(res) == di


def test_creation(ethereum_private_key_file):
    """Test the creation of the crypto_objects."""
    assert EthereumCrypto(), "Managed to initialise the eth_account"
    assert EthereumCrypto(
        ethereum_private_key_file
    ), "Managed to load the eth private key"


def test_initialization():
    """Test the initialisation of the variables."""
    account = EthereumCrypto()
    assert account.entity is not None, "The property must return the account."
    assert (
        account.address is not None and type(account.address) == str
    ), "After creation the display address must not be None"
    assert (
        account.public_key is not None and type(account.public_key) == str
    ), "After creation the public key must no be None"
    assert account.entity is not None, "After creation the entity must no be None"


def test_derive_address():
    """Test the get_address_from_public_key method"""
    account = EthereumCrypto()
    address = EthereumApi.get_address_from_public_key(account.public_key)
    assert account.address == address, "Address derivation incorrect"


def test_sign_and_recover_message(ethereum_private_key_file):
    """Test the signing and the recovery function for the eth_crypto."""
    account = EthereumCrypto(ethereum_private_key_file)
    sign_bytes = account.sign_message(message=b"hello")
    assert len(sign_bytes) > 0, "The len(signature) must not be 0"
    recovered_addresses = EthereumApi.recover_message(
        message=b"hello", signature=sign_bytes
    )
    assert len(recovered_addresses) == 1, "Wrong number of addresses recovered."
    assert (
        recovered_addresses[0] == account.address
    ), "Failed to recover the correct address."


def test_sign_and_recover_message_deprecated(ethereum_private_key_file):
    """Test the signing and the recovery function for the eth_crypto."""
    account = EthereumCrypto(ethereum_private_key_file)
    message = b"hello"
    message_hash = hashlib.sha256(message).digest()
    sign_bytes = account.sign_message(message=message_hash, is_deprecated_mode=True)
    assert len(sign_bytes) > 0, "The len(signature) must not be 0"
    recovered_addresses = EthereumApi.recover_message(
        message=message_hash, signature=sign_bytes, is_deprecated_mode=True
    )
    assert len(recovered_addresses) == 1, "Wrong number of addresses recovered."
    assert (
        recovered_addresses[0] == account.address
    ), "Failed to recover the correct address."


def test_sign_and_recover_message_public_key(ethereum_private_key_file):
    """Test the signing and the recovery function for the eth_crypto."""
    account = EthereumCrypto(ethereum_private_key_file)
    sign_bytes = account.sign_message(message=b"hello")
    assert len(sign_bytes) > 0, "The len(signature) must not be 0"
    recovered_public_keys = EthereumApi.recover_public_keys_from_message(
        message=b"hello", signature=sign_bytes
    )
    assert len(recovered_public_keys) == 1, "Wrong number of public keys recovered."
    assert (
        EthereumApi.get_address_from_public_key(recovered_public_keys[0])
        == account.address
    ), "Failed to recover the correct address."


def test_get_hash():
    """Test the get hash functionality."""
    expected_hash = "0x1c8aff950685c2ed4bc3174f3472287b56d9517b9c948127319a09a7a36deac8"
    hash_ = EthereumApi.get_hash(message=b"hello")
    assert expected_hash == hash_


def test_dump_positive(ethereum_private_key_file):
    """Test dump."""
    account = EthereumCrypto(ethereum_private_key_file)
    account.dump(MagicMock())


def test_api_creation(ethereum_testnet_config):
    """Test api instantiation."""
    assert EthereumApi(**ethereum_testnet_config), "Failed to initialise the api"


def test_api_creation_poa(ethereum_testnet_config):
    """Test api instantiation with the poa flag enabled."""
    ethereum_testnet_config["poa_chain"] = True
    assert EthereumApi(**ethereum_testnet_config), "Failed to initialise the api"


def test_api_none(ethereum_testnet_config):
    """Test the "api" of the cryptoApi is none."""
    eth_api = EthereumApi(**ethereum_testnet_config)
    assert eth_api.api is not None, "The api property is None."


def test_validate_address():
    """Test the is_valid_address functionality."""
    account = EthereumCrypto()
    assert EthereumApi.is_valid_address(account.address)
    assert not EthereumApi.is_valid_address(account.address + "wrong")


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_balance(ethereum_testnet_config, ganache, ethereum_private_key_file):
    """Test the balance is zero for a new account."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    ec = EthereumCrypto()
    balance = ethereum_api.get_balance(ec.address)
    assert balance == 0, "New account has a positive balance."
    ec = EthereumCrypto(private_key_path=ethereum_private_key_file)
    balance = ethereum_api.get_balance(ec.address)
    assert balance > 0, "Existing account has no balance."


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_state(ethereum_testnet_config, ganache):
    """Test that get_state() with 'get_block' function returns something containing the block number."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    callable_name = "get_block"
    args = ("latest",)
    block = ethereum_api.get_state(callable_name, *args)
    assert block is not None, "response to get_block is empty."
    assert "number" in block, "response to get_block() does not contain 'number'"


def _wait_get_receipt(
    ethereum_api: EthereumApi, transaction_digest: str
) -> Tuple[Optional[JSONLike], bool]:
    transaction_receipt = None
    not_settled = True
    elapsed_time = 0
    time_to_wait = 40
    sleep_time = 2
    while not_settled and elapsed_time < time_to_wait:
        elapsed_time += sleep_time
        time.sleep(sleep_time)
        transaction_receipt = ethereum_api.get_transaction_receipt(transaction_digest)
        if transaction_receipt is None:
            continue
        is_settled = ethereum_api.is_transaction_settled(transaction_receipt)
        not_settled = not is_settled

    return transaction_receipt, not not_settled


def _construct_and_settle_tx(
    ethereum_api: EthereumApi,
    account: EthereumCrypto,
    tx_params: dict,
) -> Tuple[str, JSONLike, bool]:
    """Construct and settle a transaction."""
    transfer_transaction = ethereum_api.get_transfer_transaction(**tx_params)
    assert (
        isinstance(transfer_transaction, dict) and len(transfer_transaction) == 8
    ), "Incorrect transfer_transaction constructed."

    signed_transaction = account.sign_transaction(transfer_transaction)
    assert (
        isinstance(signed_transaction, dict) and len(signed_transaction) == 5
    ), "Incorrect signed_transaction constructed."

    transaction_digest = ethereum_api.send_signed_transaction(signed_transaction)
    assert transaction_digest is not None, "Failed to submit transfer transaction!"

    transaction_receipt, is_settled = _wait_get_receipt(
        ethereum_api, transaction_digest
    )

    assert transaction_receipt is not None, "Failed to retrieve transaction receipt."

    return transaction_digest, transaction_receipt, is_settled


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_construct_sign_and_submit_transfer_transaction(
    ethereum_testnet_config, ganache, ethereum_private_key_file
):
    """Test the construction, signing and submitting of a transfer transaction."""
    account = EthereumCrypto(private_key_path=ethereum_private_key_file)
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**ethereum_testnet_config)

    tx_params = {
        "sender_address": account.address,
        "destination_address": ec2.address,
        "amount": 40000,
        "tx_fee": 30000,
        "tx_nonce": ethereum_api.generate_tx_nonce(ec2.address, account.address),
        "chain_id": DEFAULT_GANACHE_CHAIN_ID,
        "max_priority_fee_per_gas": 1_000_000_000,
        "max_fee_per_gas": 1_000_000_000,
    }

    transaction_digest, transaction_receipt, is_settled = _construct_and_settle_tx(
        ethereum_api,
        account,
        tx_params,
    )
    assert is_settled, "Failed to verify tx!"

    tx = ethereum_api.get_transaction(transaction_digest)
    is_valid = ethereum_api.is_transaction_valid(
        tx, ec2.address, account.address, tx_params["tx_nonce"], tx_params["amount"]
    )
    assert is_valid, "Failed to settle tx correctly!"
    assert tx != transaction_receipt, "Should not be same!"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_wealth_positive(caplog):
    """Test the balance is zero for a new account."""
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.ethereum._default_logger"):
        ethereum_faucet_api = EthereumFaucetApi()
        ec = EthereumCrypto()
        ethereum_faucet_api.get_wealth(ec.address, "some_url")
        assert (
            "Invalid URL" in caplog.text
        ), f"Cannot find message in output: {caplog.text}"


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_get_deploy_transaction(ethereum_testnet_config, ganache):
    """Test the get deploy transaction method."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    ec2 = EthereumCrypto()
    interface = {"abi": [], "bytecode": b""}
    max_priority_fee_per_gas = 1000000000
    max_fee_per_gas = 1000000000
    deploy_tx = ethereum_api.get_deploy_transaction(
        contract_interface=interface,
        deployer_address=ec2.address,
        value=0,
        max_priority_fee_per_gas=max_priority_fee_per_gas,
        max_fee_per_gas=max_fee_per_gas,
    )
    assert type(deploy_tx) == dict and len(deploy_tx) == 8
    assert all(
        key
        in [
            "from",
            "value",
            "gas",
            "nonce",
            "data",
            "maxPriorityFeePerGas",
            "maxFeePerGas",
            "chainId",
        ]
        for key in deploy_tx.keys()
    )


def test_load_contract_interface():
    """Test the load_contract_interface method."""
    path = Path(ROOT_DIR, "tests", "data", "dummy_contract", "build", "some.json")
    result = EthereumApi.load_contract_interface(path)
    assert "abi" in result
    assert "bytecode" in result


@patch.object(EthereumApi, "_try_get_transaction_count", return_value=None)
def test_ethereum_api_get_transfer_transaction(*args):
    """Test EthereumApi.get_transfer_transaction."""
    ec1 = EthereumCrypto()
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**get_default_gas_strategies())
    args = {
        "sender_address": ec1.address,
        "destination_address": ec2.address,
        "amount": 1,
        "tx_fee": 0,
        "tx_nonce": "",
        "max_fee_per_gas": 20,
    }
    assert ethereum_api.get_transfer_transaction(**args) is None


@patch.object(EthereumApi, "_try_get_transaction_count", return_value=1)
@patch.object(EthereumApi, "_try_get_max_priority_fee", return_value=1)
def test_ethereum_api_get_transfer_transaction_2(*args):
    """Test EthereumApi.get_transfer_transaction."""
    ec1 = EthereumCrypto()
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**get_default_gas_strategies())
    ethereum_api._is_gas_estimation_enabled = True
    args = {
        "sender_address": ec1.address,
        "destination_address": ec2.address,
        "amount": 1,
        "tx_fee": 0,
        "tx_nonce": "",
        "max_fee_per_gas": 10,
    }
    with patch.object(ethereum_api.api.eth, "estimate_gas", return_value=1):
        assert len(ethereum_api.get_transfer_transaction(**args)) == 8


@patch.object(EthereumApi, "_try_get_transaction_count", return_value=1)
def test_ethereum_api_get_transfer_transaction_3(*args):
    """Test EthereumApi.get_transfer_transaction."""
    ec1 = EthereumCrypto()
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**get_default_gas_strategies())
    ethereum_api._is_gas_estimation_enabled = True
    args = {
        "sender_address": ec1.address,
        "destination_address": ec2.address,
        "amount": 1,
        "tx_fee": 0,
        "tx_nonce": "",
        "max_fee_per_gas": 10,
    }
    with patch.object(ethereum_api.api.eth, "_max_priority_fee", return_value=1):
        assert len(ethereum_api.get_transfer_transaction(**args)) == 8


def test_ethereum_api_get_deploy_transaction(ethereum_testnet_config):
    """Test EthereumApi.get_deploy_transaction."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    ec1 = EthereumCrypto()
    with patch.object(ethereum_api.api.eth, "get_transaction_count", return_value=None):
        assert (
            ethereum_api.get_deploy_transaction(
                **{
                    "contract_interface": {"": ""},
                    "deployer_address": ec1.address,
                    "value": 1,
                    "max_fee_per_gas": 10,
                }
            )
            is None
        )

    contract_instance = Mock()
    constructor = Mock()
    constructor.build_transaction = lambda x: x
    contract_instance.constructor = Mock(return_value=constructor)

    with patch.object(
        ethereum_api.api.eth, "get_transaction_count", return_value=1
    ), patch.object(
        ethereum_api, "_is_gas_estimation_enabled", return_value=True
    ), mock.patch.object(
        ethereum_api,
        "_try_get_gas_estimate",
        return_value=120,
    ), patch.object(
        ethereum_api, "_is_gas_estimation_enabled", return_value=True
    ), patch.object(
        ethereum_api, "get_contract_instance", return_value=contract_instance
    ):
        tx = ethereum_api.get_deploy_transaction(
            **{
                "contract_interface": {"": ""},
                "deployer_address": ec1.address,
                "value": 1,
                "max_fee_per_gas": 1,
                "gas_price": 1,
                "max_priority_fee_per_gas": 1,
            }
        )
        assert tx["gas"] == 120


def test_session_cache():
    """Test session cache."""
    assert isinstance(session_cache, SimpleCacheLockWrapper)

    session_cache.cache("key", 1)
    assert session_cache.get_cache_entry("key") == 1
    session_cache.clear()
    assert "key" not in session_cache


def test_gas_price_strategy_eip1559() -> None:
    """Test eip1559 based gas price strategy."""

    callable_ = get_gas_price_strategy_eip1559(**DEFAULT_EIP1559_STRATEGY)

    web3 = Web3()
    get_block_mock = mock.patch.object(
        web3.eth, "get_block", return_value={"baseFeePerGas": 150e9, "number": 1}
    )

    fee_history_mock = mock.patch.object(
        web3.eth,
        "fee_history",
        return_value=get_history_data(
            n_blocks=5,
        ),
    )

    with get_block_mock:
        with fee_history_mock:
            gas_stregy = callable_(web3, "tx_params")

    assert all([key in gas_stregy for key in ["maxFeePerGas", "maxPriorityFeePerGas"]])
    assert all([value > 1e8 for value in gas_stregy.values()])


def test_gas_price_strategy_eip1559_estimate_none() -> None:
    """Test eip1559 based gas price strategy."""

    callable_ = get_gas_price_strategy_eip1559(**DEFAULT_EIP1559_STRATEGY)

    web3 = Web3()
    get_block_mock = mock.patch.object(
        web3.eth, "get_block", return_value={"baseFeePerGas": 150e9, "number": 1}
    )

    fee_history_mock = mock.patch.object(
        web3.eth,
        "fee_history",
        return_value=get_history_data(
            n_blocks=5,
        ),
    )
    with get_block_mock:
        with fee_history_mock:
            with mock.patch(
                "aea_ledger_ethereum.ethereum.estimate_priority_fee",
                new_callable=lambda: lambda *args, **kwargs: None,
            ):
                gas_stregy = callable_(web3, "tx_params")

    assert all([key in gas_stregy for key in ["maxFeePerGas", "maxPriorityFeePerGas"]])


def test_gas_price_strategy_eip1559_fallback() -> None:
    """Test eip1559 based gas price strategy."""

    strategy_kwargs = DEFAULT_EIP1559_STRATEGY.copy()
    strategy_kwargs["max_gas_fast"] = -1

    callable_ = get_gas_price_strategy_eip1559(**strategy_kwargs)
    web3 = Web3()
    get_block_mock = mock.patch.object(
        web3.eth, "get_block", return_value={"baseFeePerGas": 150e9, "number": 1}
    )

    fee_history_mock = mock.patch.object(
        web3.eth,
        "fee_history",
        return_value=get_history_data(
            n_blocks=5,
        ),
    )
    with get_block_mock:
        with fee_history_mock:
            with mock.patch(
                "aea_ledger_ethereum.ethereum.estimate_priority_fee",
                new_callable=lambda: lambda *args, **kwargs: None,
            ):
                gas_stregy = callable_(web3, "tx_params")

    assert all([key in gas_stregy for key in ["maxFeePerGas", "maxPriorityFeePerGas"]])


def test_gas_price_strategy_eth_gasstation():
    """Test the gas price strategy when using eth gasstation."""
    gas_price_strategy = "fast"
    excepted_result = 10
    callable_ = get_gas_price_strategy(gas_price_strategy, "api_key")
    with patch.object(
        requests,
        "get",
        return_value=MagicMock(
            status_code=200,
            json=MagicMock(return_value={gas_price_strategy: excepted_result}),
        ),
    ):
        result = callable_(Web3, "tx_params")
    assert cast(int, result["gasPrice"]) == cast(int, excepted_result / 10 * 1000000000)


def test_gas_price_strategy_not_supported(caplog):
    """Test the gas price strategy when not supported."""
    gas_price_strategy = "superfast"
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.ethereum._default_logger"):
        callable_ = get_gas_price_strategy(gas_price_strategy, "api_key")
    assert callable_ == rpc_gas_price_strategy_wrapper
    assert (
        f"Gas price strategy `{gas_price_strategy}` not in list of supported modes:"
        in caplog.text
    )


def test_gas_price_strategy_no_api_key(caplog):
    """Test the gas price strategy when no api key is provided."""
    gas_price_strategy = "fast"
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.ethereum._default_logger"):
        callable_ = get_gas_price_strategy(gas_price_strategy, None)
    assert callable_ == rpc_gas_price_strategy_wrapper
    assert (
        "No ethgasstation api key provided. Falling back to `rpc_gas_price_strategy`."
        in caplog.text
    )


def test_dump_load_with_password():
    """Test dumping and loading a key with password."""
    with tempfile.TemporaryDirectory() as dirname:
        encrypted_file_name = Path(dirname, "eth_key_encrypted")
        password = "somePwd"  # nosec
        ec = EthereumCrypto()
        ec.dump(encrypted_file_name, password)
        assert encrypted_file_name.exists()
        with pytest.raises(DecryptError, match="Decrypt error! Bad password?"):
            ec2 = EthereumCrypto.load_private_key_from_path(
                encrypted_file_name, "wrongPassw"
            )
        ec2 = EthereumCrypto(encrypted_file_name, password)
        assert ec2.private_key == ec.private_key


def test_load_errors():
    """Test load errors: bad password, no password specified."""
    ec = EthereumCrypto()
    with patch.object(EthereumCrypto, "load", return_value="bad sTring"):
        with pytest.raises(KeyIsIncorrect, match="Try to specify `password`"):
            ec.load_private_key_from_path("any path")

        with pytest.raises(KeyIsIncorrect, match="Wrong password?"):
            ec.load_private_key_from_path("any path", password="some")


def test_decrypt_error():
    """Test bad password error on decrypt."""
    ec = EthereumCrypto()
    ec._pritvate_key = EthereumCrypto.generate_private_key()
    password = "test"
    encrypted_data = ec.encrypt(password=password)
    with pytest.raises(DecryptError, match="Bad password"):
        ec.decrypt(encrypted_data, password + "some")

    with patch(
        "aea_ledger_ethereum.ethereum.Account.decrypt",
        side_effect=ValueError("expected"),
    ):
        with pytest.raises(ValueError, match="expected"):
            ec.decrypt(encrypted_data, password + "some")


def test_helper_get_contract_address():
    """Test EthereumHelper.get_contract_address."""
    assert EthereumHelper.get_contract_address({"contractAddress": "123"}) == "123"


def test_contract_method_call():
    """Test EthereumApi.contract_method_call."""

    method_mock = MagicMock()
    method_mock().call = MagicMock(return_value={"value": 0})

    contract_instance = MagicMock()
    contract_instance.functions.dummy_method = method_mock

    result = EthereumApi.contract_method_call(
        contract_instance=contract_instance, method_name="dummy_method", dummy_arg=1
    )
    assert result["value"] == 0


def test_build_transaction(ethereum_testnet_config):
    """Test EthereumApi.build_transaction."""

    def pass_tx_params(tx_params):
        return tx_params

    tx_mock = MagicMock()
    tx_mock.build_transaction = pass_tx_params

    method_mock = MagicMock(return_value=tx_mock)

    contract_instance = MagicMock()
    contract_instance.functions.dummy_method = method_mock

    eth_api = EthereumApi(**ethereum_testnet_config)

    with pytest.raises(
        ValueError, match=re.escape("Argument 'method_args' cannot be 'None'.")
    ):
        eth_api.build_transaction(
            contract_instance=contract_instance,
            method_name="dummy_method",
            method_args=None,
            tx_args={},
        )
    with pytest.raises(
        ValueError, match=re.escape("Argument 'tx_args' cannot be 'None'.")
    ):
        eth_api.build_transaction(
            contract_instance=contract_instance,
            method_name="dummy_method",
            method_args={},
            tx_args=None,
        )

    with mock.patch(
        "web3.eth.Eth.get_transaction_count",
        return_value=0,
    ):
        result = eth_api.build_transaction(
            contract_instance=contract_instance,
            method_name="dummy_method",
            method_args={},
            tx_args=dict(
                sender_address="sender_address",
                eth_value=0,
                gas=0,
                gasPrice=0,  # camel-casing due to contract api requirements
                maxFeePerGas=0,  # camel-casing due to contract api requirements
                maxPriorityFeePerGas=0,  # camel-casing due to contract api requirements
            ),
        )

        assert result == dict(
            nonce=0,
            value=0,
            gas=0,
            gasPrice=0,
            maxFeePerGas=0,
            maxPriorityFeePerGas=0,
        )

        with mock.patch.object(
            EthereumApi,
            "try_get_gas_pricing",
            return_value={"gas": 0},
        ):
            result = eth_api.build_transaction(
                contract_instance=contract_instance,
                method_name="dummy_method",
                method_args={},
                tx_args=dict(
                    sender_address="sender_address",
                    eth_value=0,
                ),
            )

            assert result == dict(nonce=0, value=0, gas=0)

        # try get gas estimates if _is_gas_estimation_enabled
        with mock.patch.object(
            EthereumApi,
            "try_get_gas_pricing",
            return_value={"gas": 0},
        ), mock.patch.object(
            eth_api,
            "_is_gas_estimation_enabled",
            return_value=True,
        ), mock.patch.object(
            eth_api,
            "_try_get_gas_estimate",
            return_value=12,
        ):
            result = eth_api.build_transaction(
                contract_instance=contract_instance,
                method_name="dummy_method",
                method_args={},
                tx_args=dict(
                    sender_address="sender_address",
                    eth_value=0,
                ),
            )

            assert result == dict(nonce=0, value=0, gas=12)


def test_get_transaction_transfer_logs(ethereum_testnet_config):
    """Test EthereumApi.get_transaction_transfer_logs."""
    eth_api = EthereumApi(**ethereum_testnet_config)

    dummy_receipt = {"logs": [{"topics": ["0x0", "0x0"]}]}

    with mock.patch(
        "web3.eth.Eth.get_transaction_receipt",
        return_value=dummy_receipt,
    ):
        contract_instance = MagicMock()
        contract_instance.events.Transfer().processReceipt.return_value = {"log": "log"}

        result = eth_api.get_transaction_transfer_logs(
            contract_instance=contract_instance,
            tx_hash="dummy_hash",
        )

        assert result == dict(logs={"log": "log"})


def test_get_transaction_transfer_logs_raise(ethereum_testnet_config):
    """Test EthereumApi.get_transaction_transfer_logs."""
    eth_api = EthereumApi(**ethereum_testnet_config)

    with mock.patch(
        "web3.eth.Eth.get_transaction_receipt",
        return_value=None,
    ):
        contract_instance = MagicMock()
        contract_instance.events.Transfer().processReceipt.return_value = {"log": "log"}

        result = eth_api.get_transaction_transfer_logs(
            contract_instance=contract_instance,
            tx_hash="dummy_hash",
        )

        assert result == dict(logs=[])


@pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
@pytest.mark.integration
@pytest.mark.ledger
def test_revert_reason(
    ethereum_private_key_file: str,
    ethereum_testnet_config: dict,
    ganache: Generator,
) -> None:
    """Test the retrieval of the revert reason for a transaction."""
    account = EthereumCrypto(private_key_path=ethereum_private_key_file)
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**ethereum_testnet_config)

    tx_params = {
        "sender_address": account.address,
        "destination_address": ec2.address,
        "amount": 40000,
        "tx_fee": 30000,
        "tx_nonce": 0,
        "chain_id": DEFAULT_GANACHE_CHAIN_ID,
        "max_priority_fee_per_gas": 1_000_000_000,
        "max_fee_per_gas": 1_000_000_000,
    }

    with mock.patch(
        "web3.eth.Eth.get_transaction_receipt",
        return_value=AttributeDict({"status": 0}),
    ):
        with mock.patch(
            "web3.eth.Eth.call",
            side_effect=ContractLogicError("test revert reason"),
        ):
            _, transaction_receipt, is_settled = _construct_and_settle_tx(
                ethereum_api,
                account,
                tx_params,
            )

            assert transaction_receipt["revert_reason"] == "test revert reason"


@pytest.mark.parametrize(
    "strategy",
    (
        {"name": EIP1559, "params": ("maxPriorityFeePerGas", "maxFeePerGas")},
        {"name": GAS_STATION, "params": ("gasPrice",)},
        {"name": EIP1559_POLYGON, "params": ("maxPriorityFeePerGas", "maxFeePerGas")},
    ),
)
def test_try_get_gas_pricing(
    strategy: Dict[str, Union[str, Tuple[str, ...]]],
    ethereum_testnet_config: dict,
    ganache: Generator,
) -> None:
    """Test `try_get_gas_pricing`."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)

    # test gas pricing
    gas_price = ethereum_api.try_get_gas_pricing(gas_price_strategy=strategy["name"])
    assert set(strategy["params"]) == set(gas_price.keys())
    assert all(
        gas_price[param] > 0 and isinstance(gas_price[param], int)
        for param in strategy["params"]
    )
    expteced_reprice = {
        param: math.ceil(value * TIP_INCREASE) for param, value in gas_price.items()
    }

    # test gas repricing
    gas_reprice = ethereum_api.try_get_gas_pricing(
        gas_price_strategy=strategy["name"], old_price=gas_price
    )
    assert all(
        gas_reprice[param] > 0 and isinstance(gas_reprice[param], int)
        for param in strategy["params"]
    )
    assert gas_reprice == expteced_reprice, "The repricing was performed incorrectly!"


@pytest.mark.parametrize(
    "strategy",
    ({"name": EIP1559_POLYGON, "params": ("maxPriorityFeePerGas", "maxFeePerGas")},),
)
def test_try_get_gas_pricing_poa(
    strategy: Dict[str, Union[str, Tuple[str, ...]]],
    polygon_testnet_config: dict,
    ganache: Generator,
) -> None:
    """Test `try_get_gas_pricing` for a poa chain like Rinkeby."""
    ethereum_api = EthereumApi(**polygon_testnet_config)
    assert "geth_poa_middleware" in ethereum_api.api.middleware_onion.keys()

    # test gas pricing
    gas_price = ethereum_api.try_get_gas_pricing(gas_price_strategy=strategy["name"])
    assert set(strategy["params"]) == set(gas_price.keys())
    assert all(
        gas_price[param] > 0 and isinstance(gas_price[param], int)
        for param in strategy["params"]
    )
    expteced_reprice = {
        param: math.ceil(value * TIP_INCREASE) for param, value in gas_price.items()
    }

    # test gas repricing
    gas_reprice = ethereum_api.try_get_gas_pricing(
        gas_price_strategy=strategy["name"], old_price=copy.deepcopy(gas_price)
    )
    assert all(
        gas_reprice[param] > 0 and isinstance(gas_reprice[param], int)
        for param in strategy["params"]
    )
    assert gas_reprice == expteced_reprice, "The repricing was performed incorrectly!"


@pytest.mark.parametrize("mock_exception", (True, False))
def test_gas_estimation(
    mock_exception,
    ethereum_testnet_config: dict,
    ganache: Generator,
    caplog,
) -> None:
    """Test gas estimation."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)
    tx = {
        "value": 0,
        "chainId": 1337,
        "from": "0xBcd4042DE499D14e55001CcbB24a551F3b954096",
        "gas": 291661,
        "maxPriorityFeePerGas": 3000000000,
        "maxFeePerGas": 4000000000,
        "to": "0x68FCdF52066CcE5612827E872c45767E5a1f6551",
        "data": "",
    }
    with caplog.at_level(logging.DEBUG, logger="aea.crypto.ethereum._default_logger"):
        with patch.object(ethereum_api._api.eth, "estimate_gas") as estimate_gas_mock:
            if mock_exception:
                # raise exception on first call only
                estimate_gas_mock.side_effect = [
                    ValueError("triggered exception"),
                    None,
                ]
            ethereum_api.update_with_gas_estimate(tx)
        if mock_exception:
            assert (
                "ValueError: triggered exception" in caplog.text
            ), f"Cannot find message in output: {caplog.text}"


@patch.object(EthereumApi, "_try_get_transaction_count", return_value=1)
@patch.object(EthereumApi, "_try_get_max_priority_fee", return_value=1)
def test_ethereum_api_get_transfer_transaction_estimate_gas(*args) -> None:
    """Test EthereumApi.get_transfer_transaction gas auto estimate."""
    ec1 = EthereumCrypto()
    ec2 = EthereumCrypto()
    ethereum_api = EthereumApi(**get_default_gas_strategies())
    ethereum_api._is_gas_estimation_enabled = True
    args = {
        "sender_address": ec1.address,
        "destination_address": ec2.address,
        "amount": 1,
        "tx_fee": 0,
        "tx_nonce": "",
        "max_fee_per_gas": None,
    }
    with patch.object(
        ethereum_api, "try_get_gas_pricing", return_value={"gasPrice": 12}
    ) as get_gas_pricing:
        result = ethereum_api.get_transfer_transaction(**args)
        assert result["gasPrice"] == 12
        get_gas_pricing.assert_called_once()

    args["gas_price"] = 13
    result = ethereum_api.get_transfer_transaction(**args)
    assert result["gasPrice"] == 13


def test_get_base_fee_multiplier() -> None:
    """Test get_base_fee_multiplier."""
    assert get_base_fee_multiplier(35) == 2.0
    assert get_base_fee_multiplier(45) == 1.6
    assert get_base_fee_multiplier(105) == 1.4
    assert get_base_fee_multiplier(205) == 1.2


def test_estimate_priority_fee() -> None:
    """Test estimate_priority_fee."""
    # return none on no rewards
    web3_mock = Mock()
    web3_mock.eth.fee_history = Mock(return_value={"reward": []})
    assert estimate_priority_fee(web3_mock, 1, 1, 1, 1, 11, 1, 1) is None

    # If we have big increase in value, we could be considering "outliers" in our estimate
    # Skip the low elements and take a new median
    web3_mock.eth.fee_history = Mock(return_value={"reward": [[1], [10], [10000]]})
    assert estimate_priority_fee(web3_mock, 1, 1, 1, 1, 11, 1, 1) == 10000


def test_try_get_revert_reason_http_error_propagated(ethereum_testnet_config) -> None:
    """Test httperror reraised if get_revert_reason fails."""
    eth_api = EthereumApi(**ethereum_testnet_config)
    tx = {"from": 1, "to": 1, "input": 1, "value": 1, "blockNumber": 1}
    api_mock = Mock()
    api_mock.eth.call = Mock(side_effect=HTTPError("http_error"))
    with patch.object(eth_api, "_api", new=api_mock):
        with pytest.raises(HTTPError):
            eth_api._try_get_revert_reason(tx, raise_on_try=True)


def test_get_gas_price_strategy() -> None:
    """Test get_gas_price_strategy and strategies."""
    strategy = get_gas_price_strategy(None)
    assert strategy is rpc_gas_price_strategy_wrapper
    assert "gasPrice" in strategy(Mock(), Mock())

    strategy = get_gas_price_strategy_eip1559_polygon("test", {"gasPrice": 12})

    resp_mock = Mock()
    resp_mock.status_code = 300
    with patch("aea.helpers.http_requests.get", return_value=resp_mock):
        assert strategy(Mock(), Mock()) == {"gasPrice": 12}

    resp_mock.status_code = 200
    resp_mock.json = Mock(return_value={"fast": {"maxFee": 2, "maxPriorityFee": 2}})
    with patch("aea.helpers.http_requests.get", return_value=resp_mock):
        assert strategy(Mock(), Mock()) == {
            "maxFeePerGas": 2000000000,
            "maxPriorityFeePerGas": 2000000000,
        }

    with patch(
        "aea.helpers.http_requests.get",
        side_effect=requests.exceptions.RequestException(Mock()),
    ):
        assert strategy(Mock(), Mock()) == {"gasPrice": 12}


@pytest.mark.ledger
def test_update_with_gas_estimate_method_raise_on_try(
    ethereum_testnet_config, ganache, ethereum_private_key_file, caplog
):
    """Test the balance is zero for a new account."""
    ethereum_api = EthereumApi(**ethereum_testnet_config)

    # The exception will be intercepted and logged
    with mock.patch.object(
        ethereum_api.api.eth, "estimate_gas", side_effect=ContractLogicError
    ), caplog.at_level(logging.WARNING):
        ethereum_api.update_with_gas_estimate(
            transaction={
                "value": 0,
                "chainId": 1337,
                "from": "0xBcd4042DE499D14e55001CcbB24a551F3b954096",
                "gas": 291661,
                "maxPriorityFeePerGas": 3000000000,
                "maxFeePerGas": 4000000000,
                "to": "0x68FCdF52066CcE5612827E872c45767E5a1f6551",
                "data": "",
            }
        )

        assert (
            "Unable to estimate gas with default state , ContractLogicError"
            in caplog.text
        )

    # The exception won't be intercepted
    with pytest.raises(ContractLogicError):
        with mock.patch.object(
            ethereum_api.api.eth, "estimate_gas", side_effect=ContractLogicError
        ):
            ethereum_api.update_with_gas_estimate(
                transaction={
                    "value": 0,
                    "chainId": 1337,
                    "from": "0xBcd4042DE499D14e55001CcbB24a551F3b954096",
                    "gas": 291661,
                    "maxPriorityFeePerGas": 3000000000,
                    "maxFeePerGas": 4000000000,
                    "to": "0x68FCdF52066CcE5612827E872c45767E5a1f6551",
                    "data": "",
                },
                raise_on_try=True,
            )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: oef_search.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10oef_search.proto\x12\x1d\x61\x65\x61.fetchai.oef_search.v1_0_0"\x89\r\n\x10OefSearchMessage\x12[\n\toef_error\x18\x05 \x01(\x0b\x32\x46.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Oef_Error_PerformativeH\x00\x12i\n\x10register_service\x18\x06 \x01(\x0b\x32M.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Register_Service_PerformativeH\x00\x12\x63\n\rsearch_result\x18\x07 \x01(\x0b\x32J.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Search_Result_PerformativeH\x00\x12g\n\x0fsearch_services\x18\x08 \x01(\x0b\x32L.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Search_Services_PerformativeH\x00\x12W\n\x07success\x18\t \x01(\x0b\x32\x44.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Success_PerformativeH\x00\x12m\n\x12unregister_service\x18\n \x01(\x0b\x32O.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Unregister_Service_PerformativeH\x00\x1a!\n\nAgentsInfo\x12\x13\n\x0b\x61gents_info\x18\x01 \x01(\x0c\x1a(\n\x0b\x44\x65scription\x12\x19\n\x11\x64\x65scription_bytes\x18\x01 \x01(\x0c\x1a\xdb\x01\n\x11OefErrorOperation\x12\x61\n\toef_error\x18\x01 \x01(\x0e\x32N.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.OefErrorOperation.OefErrorEnum"c\n\x0cOefErrorEnum\x12\x14\n\x10REGISTER_SERVICE\x10\x00\x12\x16\n\x12UNREGISTER_SERVICE\x10\x01\x12\x13\n\x0fSEARCH_SERVICES\x10\x02\x12\x10\n\x0cSEND_MESSAGE\x10\x03\x1a\x1c\n\x05Query\x12\x13\n\x0bquery_bytes\x18\x01 \x01(\x0c\x1ay\n\x1dRegister_Service_Performative\x12X\n\x13service_description\x18\x01 \x01(\x0b\x32;.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Description\x1a{\n\x1fUnregister_Service_Performative\x12X\n\x13service_description\x18\x01 \x01(\x0b\x32;.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Description\x1a\x64\n\x1cSearch_Services_Performative\x12\x44\n\x05query\x18\x01 \x01(\x0b\x32\x35.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.Query\x1a}\n\x1aSearch_Result_Performative\x12\x0e\n\x06\x61gents\x18\x01 \x03(\t\x12O\n\x0b\x61gents_info\x18\x02 \x01(\x0b\x32:.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.AgentsInfo\x1ag\n\x14Success_Performative\x12O\n\x0b\x61gents_info\x18\x01 \x01(\x0b\x32:.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.AgentsInfo\x1ax\n\x16Oef_Error_Performative\x12^\n\x13oef_error_operation\x18\x01 \x01(\x0b\x32\x41.aea.fetchai.oef_search.v1_0_0.OefSearchMessage.OefErrorOperationB\x0e\n\x0cperformativeb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "oef_search_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _globals["_OEFSEARCHMESSAGE"]._serialized_start = 52
    _globals["_OEFSEARCHMESSAGE"]._serialized_end = 1725
    _globals["_OEFSEARCHMESSAGE_AGENTSINFO"]._serialized_start = 678
    _globals["_OEFSEARCHMESSAGE_AGENTSINFO"]._serialized_end = 711
    _globals["_OEFSEARCHMESSAGE_DESCRIPTION"]._serialized_start = 713
    _globals["_OEFSEARCHMESSAGE_DESCRIPTION"]._serialized_end = 753
    _globals["_OEFSEARCHMESSAGE_OEFERROROPERATION"]._serialized_start = 756
    _globals["_OEFSEARCHMESSAGE_OEFERROROPERATION"]._serialized_end = 975
    _globals["_OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM"]._serialized_start = 876
    _globals["_OEFSEARCHMESSAGE_OEFERROROPERATION_OEFERRORENUM"]._serialized_end = 975
    _globals["_OEFSEARCHMESSAGE_QUERY"]._serialized_start = 977
    _globals["_OEFSEARCHMESSAGE_QUERY"]._serialized_end = 1005
    _globals["_OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE"]._serialized_start = 1007
    _globals["_OEFSEARCHMESSAGE_REGISTER_SERVICE_PERFORMATIVE"]._serialized_end = 1128
    _globals[
        "_OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE"
    ]._serialized_start = 1130
    _globals["_OEFSEARCHMESSAGE_UNREGISTER_SERVICE_PERFORMATIVE"]._serialized_end = 1253
    _globals["_OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE"]._serialized_start = 1255
    _globals["_OEFSEARCHMESSAGE_SEARCH_SERVICES_PERFORMATIVE"]._serialized_end = 1355
    _globals["_OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE"]._serialized_start = 1357
    _globals["_OEFSEARCHMESSAGE_SEARCH_RESULT_PERFORMATIVE"]._serialized_end = 1482
    _globals["_OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE"]._serialized_start = 1484
    _globals["_OEFSEARCHMESSAGE_SUCCESS_PERFORMATIVE"]._serialized_end = 1587
    _globals["_OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE"]._serialized_start = 1589
    _globals["_OEFSEARCHMESSAGE_OEF_ERROR_PERFORMATIVE"]._serialized_end = 1709
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test protocol helpers."""

from pathlib import Path

import pytest
import yaml

from aea.configurations.base import PublicId
from aea.helpers.protocols import (
    get_protocol_specification_from_readme,
    get_protocol_specification_id_from_specification,
)


@pytest.mark.parametrize("protocol_yaml_path", Path("packages").rglob("protocol.yaml"))
def test_get_protocol_specification_from_readme(protocol_yaml_path):
    """Test get_protocol_specification_from_readme"""

    spec = get_protocol_specification_from_readme(protocol_yaml_path.parent)
    assert all(yaml.safe_load_all(spec))
    spec_id = get_protocol_specification_id_from_specification(spec)
    assert PublicId.from_str(spec_id)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains the tests for the `aea scaffold` sub-command."""

from pathlib import Path


def files_outside_copyright_are_identical(*files: Path) -> bool:
    """Check files are identical outside copyright author / year"""

    def remove_copyright_author_year_lines(s: str) -> str:
        """Filter copyright author and year for file comparison"""
        lines, prefix = s.splitlines(), "#   Copyright"
        return "/n".join(line for line in lines if not line.startswith(prefix))

    lines = (f.read_text() for f in files)
    cleaned_lines = set(map(remove_copyright_author_year_lines, lines))
    return len(set(cleaned_lines)) == 1



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'aea launch' subcommand."""
import os
import sys
from collections import OrderedDict
from pathlib import Path
from typing import List, Optional, cast

import click

from aea.cli.utils.click_utils import AgentDirectory, password_option
from aea.cli.utils.context import Context
from aea.cli.utils.loggers import logger
from aea.helpers.multiple_executor import ExecutorExceptionPolicies
from aea.launcher import AEALauncher


@click.command()
@click.argument("agents", nargs=-1, type=AgentDirectory())
@password_option()
@click.option("--multithreaded", is_flag=True)
@click.pass_context
def launch(
    click_context: click.Context,
    agents: List[str],
    password: Optional[str],
    multithreaded: bool,
) -> None:
    """Launch many agents at the same time."""
    _launch_agents(click_context, agents, multithreaded, password)


def _launch_agents(
    click_context: click.core.Context,
    agents: List[str],
    multithreaded: bool,
    password: Optional[str] = None,
) -> None:
    """
    Run multiple agents.

    :param click_context: click context object.
    :param agents: agents names.
    :param multithreaded: bool flag to run as multithreads.
    :param password: the password to encrypt/decrypt the private key.
    """
    agents_directories = list(map(Path, list(OrderedDict.fromkeys(agents))))
    mode = "threaded" if multithreaded else "multiprocess"
    ctx = cast(Context, click_context.obj)

    launcher = AEALauncher(
        agent_dirs=agents_directories,
        mode=mode,
        fail_policy=ExecutorExceptionPolicies.log_only,
        log_level=ctx.verbosity,
        password=password,
    )

    try:
        """
        run in threaded mode and wait for thread finished cause issue with python 3.7 on windows
        probably keyboard interrupt exception gets lost in executor pool or in asyncio module
        """
        launcher.start(threaded=True)
        launcher.try_join_thread()
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt detected.")
    finally:
        timeout: Optional[float] = None
        if os.name == "nt":
            # Windows bug: https://bugs.python.org/issue21822
            timeout = 0  # pragma: nocover
        launcher.stop(timeout)

    for agent in launcher.failed:
        logger.info(f"Agent {agent} terminated with exit code 1")

    for agent in launcher.not_failed:
        logger.info(f"Agent {agent} terminated with exit code 0")

    logger.debug(f"Exit cli. code: {launcher.num_failed}")
    sys.exit(1 if launcher.num_failed > 0 else 0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains testing utilities."""
import logging
import re
import shutil
import subprocess  # nosec
import time
from abc import ABC, abstractmethod
from typing import Any, Generator, cast

import pytest


try:
    import docker
    from docker import DockerClient
    from docker.models.containers import Container
except ImportError:  # pragma: no cover
    # to avoid having to make docker a framework dependency
    docker = DockerClient = Container = Any


logger = logging.getLogger(__name__)


class DockerImage(ABC):
    """A class to wrap interaction with a Docker image."""

    MINIMUM_DOCKER_VERSION = (19, 0, 0)

    def __init__(self, client: DockerClient):
        """Initialize."""
        self._client = client

    def check_skip(self) -> None:
        """
        Check whether the test should be skipped.

        By default, nothing happens.
        """
        self._check_docker_binary_available()

    def _check_docker_binary_available(self) -> None:
        """Check the 'Docker' CLI tool is in the OS PATH."""
        result = shutil.which("docker")
        if result is None:
            pytest.skip("Docker not in the OS Path; skipping the test")
            return  # pragma: no cover

        proc_result = subprocess.run(  # pylint: disable=subprocess-run-check # nosec
            ["docker", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if proc_result.returncode != 0:
            pytest.skip(
                f"'docker --version' failed with exit code {proc_result.returncode}"
            )
            return  # pragma: no cover

        match = re.search(
            r"Docker version ([0-9]+)\.([0-9]+)\.([0-9]+)",
            proc_result.stdout.decode("utf-8"),
        )
        if match is None:
            pytest.skip("cannot read version from the output of 'docker --version'")
            return  # pragma: no cover

        match = cast(re.Match, match)
        version = (int(match.group(1)), int(match.group(2)), int(match.group(3)))
        if version < self.MINIMUM_DOCKER_VERSION:
            pytest.skip(
                f"expected Docker version to be at least {'.'.join([str(item) for item in self.MINIMUM_DOCKER_VERSION])}, found {'.'.join([str(item) for item in version])}"
            )
            return  # pragma: no cover

    @property
    @abstractmethod
    def tag(self) -> str:
        """Return the tag of the image."""

    def stop_if_already_running(self) -> None:
        """Stop the running images with the same tag, if any."""

        client = docker.from_env()
        for container in client.containers.list():
            if self.tag in container.image.tags:
                logger.info(f"Stopping image {self.tag}...")
                container.stop()

    @abstractmethod
    def create(self) -> Container:
        """Instantiate the image in a container."""

    @abstractmethod
    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """
        Wait until the image is running.

        :param max_attempts: max number of attempts.
        :param sleep_rate: the amount of time to sleep between different requests.
        :return: True if the wait was successful, False otherwise.
        """
        return True  # pragma: no cover


def launch_image(
    image: DockerImage, timeout: float = 2.0, max_attempts: int = 10
) -> Generator:
    """Launch image."""

    image.check_skip()
    image.stop_if_already_running()
    container = image.create()
    container.start()
    logger.info(f"Setting up image {image.tag}...")
    success = image.wait(max_attempts, timeout)
    if not success:
        container.stop()
        container.remove()
        pytest.fail(f"{image.tag} doesn't work. Exiting...")

    try:
        logger.info("Done!")
        time.sleep(timeout)
        yield
    finally:
        logger.info(f"Stopping the image {image.tag}...")
        container.stop()
        container.remove()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 fetchai
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for oef_search protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.helpers.search.models import Constraint, ConstraintType
from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.fetchai.protocols.oef_search.custom_types import (
    AgentsInfo,
    Description,
    Query,
)
from packages.fetchai.protocols.oef_search.message import OefSearchMessage


class TestMessageOefSearch(BaseProtocolMessagesTestCase):
    """Test for the 'oef_search' protocol message."""

    MESSAGE_CLASS = OefSearchMessage

    def build_messages(self) -> List[OefSearchMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            OefSearchMessage(
                performative=OefSearchMessage.Performative.REGISTER_SERVICE,
                service_description=Description(
                    {"foo1": 1, "bar1": 2}
                ),  # check it please!
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
                service_description=Description(
                    {"foo1": 1, "bar1": 2}
                ),  # check it please!
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.SEARCH_SERVICES,
                query=Query(
                    [Constraint("something", ConstraintType(">", 1))]
                ),  # check it please!
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                agents=("some str",),
                agents_info=AgentsInfo(
                    {
                        "key_1": {"key_1": b"value_1", "key_2": b"value_2"},
                        "key_2": {"key_3": b"value_3", "key_4": b"value_4"},
                    }
                ),
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.SUCCESS,
                agents_info=AgentsInfo(
                    {
                        "key_1": {"key_1": b"value_1", "key_2": b"value_2"},
                        "key_2": {"key_3": b"value_3", "key_4": b"value_4"},
                    }
                ),
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.OEF_ERROR,
                oef_error_operation=OefSearchMessage.OefErrorOperation.OTHER,  # check it please!
            ),
        ]

    def build_inconsistent(self) -> List[OefSearchMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            OefSearchMessage(
                performative=OefSearchMessage.Performative.REGISTER_SERVICE,
                # skip content: service_description
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.UNREGISTER_SERVICE,
                # skip content: service_description
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.SEARCH_SERVICES,
                # skip content: query
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                # skip content: agents
                agents_info=AgentsInfo(
                    {
                        "key_1": {"key_1": b"value_1", "key_2": b"value_2"},
                        "key_2": {"key_3": b"value_3", "key_4": b"value_4"},
                    }
                ),  # check it please!
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.SUCCESS,
                # skip content: agents_info
            ),
            OefSearchMessage(
                performative=OefSearchMessage.Performative.OEF_ERROR,
                # skip content: oef_error_operation
            ),
        ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the erc1155_deploy skill."""
# pylint: skip-file

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.generic import (
    AGENT_LOCATION_MODEL,
    AGENT_PERSONALITY_MODEL,
    AGENT_REMOVE_SERVICE_MODEL,
    AGENT_SET_SERVICE_MODEL,
)
from aea.helpers.search.models import Description, Location
from aea.helpers.transaction.base import Terms

from packages.fetchai.contracts.erc1155.contract import ERC1155Contract
from packages.fetchai.skills.erc1155_deploy.strategy import (
    SIMPLE_SERVICE_MODEL,
    Strategy,
)
from packages.fetchai.skills.erc1155_deploy.tests.intermediate_class import (
    ERC1155DeployTestCase,
)


class TestStrategy(ERC1155DeployTestCase):
    """Test Strategy of erc1155_deploy."""

    def test__init__(self):
        """Test the properties of Strategy class."""
        assert Strategy(
            location=self.location,
            mint_quantities=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            service_data=self.service_data,
            personality_data=self.personality_data,
            classification=self.classification,
            from_supply=self.from_supply,
            to_supply=self.to_supply,
            value=self.value,
            token_type=1,
            name="strategy",
            skill_context=self.skill.skill_context,
        )

    def test_properties(self):
        """Test the properties of Strategy class."""
        assert self.strategy.ledger_id == self.skill.skill_context.default_ledger_id
        assert self.strategy.contract_id == str(ERC1155Contract.contract_id)
        assert self.strategy.mint_quantities == self.mint_quantities
        assert self.strategy.token_ids == self.strategy._token_ids

        self.strategy._token_ids = None
        with pytest.raises(ValueError, match="Token ids not set."):
            assert self.strategy.token_ids

        with pytest.raises(ValueError, match="Contract address not set!"):
            assert self.strategy.contract_address
        self.strategy.contract_address = self.contract_address
        with pytest.raises(AEAEnforceError, match="Contract address already set!"):
            self.strategy.contract_address = self.contract_address
        assert self.strategy.contract_address == self.contract_address

        assert self.strategy.is_contract_deployed is False
        self.strategy.is_contract_deployed = True
        assert self.strategy.is_contract_deployed is True
        with pytest.raises(AEAEnforceError, match="Only allowed to switch to true."):
            self.strategy.is_contract_deployed = False

        assert self.strategy.is_tokens_created is False
        self.strategy.is_tokens_created = True
        assert self.strategy.is_tokens_created is True
        with pytest.raises(AEAEnforceError, match="Only allowed to switch to true."):
            self.strategy.is_tokens_created = False

        assert self.strategy.is_tokens_minted is False
        self.strategy.is_tokens_minted = True
        assert self.strategy.is_tokens_minted is True
        with pytest.raises(AEAEnforceError, match="Only allowed to switch to true."):
            self.strategy.is_tokens_minted = False

        assert self.strategy.gas == self.strategy._gas

    def test_get_location_description(self):
        """Test the get_location_description method of the Strategy class."""
        description = self.strategy.get_location_description()

        assert type(description) == Description
        assert description.data_model is AGENT_LOCATION_MODEL
        assert description.values.get("location", "") == Location(
            latitude=self.location["latitude"], longitude=self.location["longitude"]
        )

    def test_get_register_service_description(self):
        """Test the get_register_service_description method of the Strategy class."""
        description = self.strategy.get_register_service_description()

        assert type(description) == Description
        assert description.data_model is AGENT_SET_SERVICE_MODEL
        assert description.values.get("key", "") == self.service_data["key"]
        assert description.values.get("value", "") == self.service_data["value"]

    def test_get_register_personality_description(self):
        """Test the get_register_personality_description method of the Strategy class."""
        description = self.strategy.get_register_personality_description()

        assert type(description) == Description
        assert description.data_model is AGENT_PERSONALITY_MODEL
        assert description.values.get("piece", "") == self.personality_data["piece"]
        assert description.values.get("value", "") == self.personality_data["value"]

    def test_get_register_classification_description(self):
        """Test the get_register_classification_description method of the Strategy class."""
        description = self.strategy.get_register_classification_description()

        assert type(description) == Description
        assert description.data_model is AGENT_PERSONALITY_MODEL
        assert description.values.get("piece", "") == self.classification["piece"]
        assert description.values.get("value", "") == self.classification["value"]

    def test_get_service_description(self):
        """Test the get_service_description method of the Strategy class."""
        description = self.strategy.get_service_description()

        assert type(description) == Description
        assert description.data_model is SIMPLE_SERVICE_MODEL
        assert (
            description.values.get("seller_service", "") == self.service_data["value"]
        )

    def test_get_unregister_service_description(self):
        """Test the get_unregister_service_description method of the Strategy class."""
        description = self.strategy.get_unregister_service_description()

        assert type(description) == Description
        assert description.data_model is AGENT_REMOVE_SERVICE_MODEL
        assert description.values.get("key", "") == self.service_data["key"]

    def test_get_deploy_terms(self):
        """Test the get_deploy_terms of Strategy."""
        assert self.strategy.get_deploy_terms() == Terms(
            ledger_id=self.ledger_id,
            sender_address=self.skill.skill_context.agent_address,
            counterparty_address=self.skill.skill_context.agent_address,
            amount_by_currency_id={},
            quantities_by_good_id={},
            nonce="",
        )

    def test_get_create_token_terms(self):
        """Test the get_create_token_terms of Parameters."""
        assert self.strategy.get_create_token_terms() == Terms(
            ledger_id=self.ledger_id,
            sender_address=self.skill.skill_context.agent_address,
            counterparty_address=self.skill.skill_context.agent_address,
            amount_by_currency_id={},
            quantities_by_good_id={},
            nonce="",
        )

    def test_get_mint_token_terms(self):
        """Test the get_mint_token_terms of Strategy."""
        assert self.strategy.get_mint_token_terms() == Terms(
            ledger_id=self.ledger_id,
            sender_address=self.skill.skill_context.agent_address,
            counterparty_address=self.skill.skill_context.agent_address,
            amount_by_currency_id={},
            quantities_by_good_id={},
            nonce="",
        )

    def test_get_proposal(self):
        """Test the get_proposal of Strategy."""
        # setup
        self.strategy._contract_address = self.contract_address
        first_id = 8768
        self.strategy._token_ids = [first_id, 234, 879643]

        # operation
        actual_proposal = self.strategy.get_proposal()

        # after
        assert all(
            keys in actual_proposal.values
            for keys in [
                "contract_address",
                "token_id",
                "trade_nonce",
                "from_supply",
                "to_supply",
            ]
        )
        assert (
            actual_proposal.values.get("contract_address", "") == self.contract_address
        )
        assert actual_proposal.values.get("token_id", "") == str(first_id)
        assert isinstance(actual_proposal.values.get("trade_nonce", ""), str)
        assert actual_proposal.values.get("from_supply", "") == str(self.from_supply)
        assert actual_proposal.values.get("to_supply", "") == str(self.to_supply)
        assert actual_proposal.values.get("value", "") == str(self.value)

    def test_get_single_swap_terms(self):
        """Test the get_single_swap_terms of Strategy."""
        assert self.strategy.get_single_swap_terms(
            self.mocked_proposal, "some_address"
        ) == Terms(
            ledger_id=self.ledger_id,
            sender_address=self.skill.skill_context.agent_address,
            counterparty_address="some_address",
            amount_by_currency_id={
                str(self.mocked_proposal.values["token_id"]): int(
                    self.mocked_proposal.values["from_supply"]
                )
                - int(self.mocked_proposal.values["to_supply"])
            },
            quantities_by_good_id={},
            is_sender_payable_tx_fee=True,
            nonce=str(self.mocked_proposal.values["trade_nonce"]),
            fee_by_currency_id={},
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Check amount of time for acn connection communications."""
import asyncio
import logging
import os
import time
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from typing import Callable, List, Tuple, Union

from aea_cli_benchmark.case_acn_communication.utils import (
    DEFAULT_DELEGATE_PORT,
    DEFAULT_MAILBOX_PORT,
    DEFAULT_NODE_PORT,
    _make_libp2p_client_connection,
    _make_libp2p_connection,
    _make_libp2p_mailbox_connection,
)

from aea.connections.base import Connection
from aea.mail.base import Envelope

from packages.fetchai.protocols.default.message import DefaultMessage


class TimeMeasure:
    """Time measure data class."""

    def __init__(self):
        """Init data class instance."""
        self.time = -1


@contextmanager
def time_measure():
    """Get time measure context."""
    start = time.time()
    m = TimeMeasure()
    try:
        yield m
    finally:
        m.time = time.time() - start


def make_envelope(from_addr: str, to_addr: str) -> Envelope:
    """Construct an envelope."""
    msg = DefaultMessage(
        dialogue_reference=("", ""),
        message_id=1,
        target=0,
        performative=DefaultMessage.Performative.BYTES,
        content=b"hello",
    )
    envelope = Envelope(
        to=to_addr,
        sender=from_addr,
        message=msg,
    )
    return envelope


async def _run(con_maker: Callable[..., Connection]) -> Tuple[float, float]:
    """Run test case and return times for the first and the second messages sent over ACN."""
    try:
        connections = []
        genesis_node = _make_libp2p_connection(".", relay=True)
        await genesis_node.connect()
        connections.append(genesis_node)
        genesis_multiaddr = genesis_node.node.multiaddrs[0]

        relay_node1 = _make_libp2p_connection(
            ".",
            relay=True,
            entry_peers=[genesis_multiaddr],
            port=DEFAULT_NODE_PORT + 1,
            mailbox=True,
            delegate=True,
            mailbox_port=DEFAULT_MAILBOX_PORT + 1,
            delegate_port=DEFAULT_DELEGATE_PORT + 1,
        )
        await relay_node1.connect()
        connections.append(relay_node1)
        relay_node2 = _make_libp2p_connection(
            ".",
            relay=True,
            entry_peers=[genesis_multiaddr],
            port=DEFAULT_NODE_PORT + 2,
            mailbox=True,
            delegate=True,
            mailbox_port=DEFAULT_MAILBOX_PORT + 2,
            delegate_port=DEFAULT_DELEGATE_PORT + 2,
        )
        await relay_node2.connect()
        connections.append(relay_node2)

        relay_node1_multiaddr = relay_node1.node.multiaddrs[0]
        relay_node2_multiaddr = relay_node2.node.multiaddrs[0]
        await asyncio.sleep(1)
        con1 = con_maker(
            port=DEFAULT_NODE_PORT + 10,
            entry_peer=relay_node1_multiaddr,
            mailbox_port=DEFAULT_MAILBOX_PORT + 1,
            delegate_port=DEFAULT_DELEGATE_PORT + 1,
            pub_key=relay_node1.node.pub,
        )
        await con1.connect()
        connections.append(con1)

        con2 = con_maker(
            port=DEFAULT_NODE_PORT + 20,
            entry_peer=relay_node2_multiaddr,
            mailbox_port=DEFAULT_MAILBOX_PORT + 2,
            delegate_port=DEFAULT_DELEGATE_PORT + 2,
            pub_key=relay_node2.node.pub,
        )
        await con2.connect()
        connections.append(con2)

        envelope = make_envelope(con1.address, con2.address)

        with time_measure() as tm:
            await con1.send(envelope)
            envelope = await con2.receive()
        first_time = tm.time

        with time_measure() as tm:
            await con1.send(envelope)
            envelope = await con2.receive()

        second_time = tm.time

        return first_time, second_time

    finally:
        for con in reversed(connections):
            await con.disconnect()


def run(connection: str, run_times: int = 10) -> List[Tuple[str, Union[int, float]]]:
    """Check construction time and memory usage."""
    logging.basicConfig(level=logging.CRITICAL)
    cwd = os.getcwd()
    try:
        if connection == "p2pnode":

            def con_maker(
                port: int,
                entry_peer: str,
                mailbox_port: int,
                delegate_port: int,
                pub_key: str,
            ):
                return _make_libp2p_connection(".", port=port, entry_peers=[entry_peer])

        elif connection == "client":

            def con_maker(
                port: int,
                entry_peer: str,
                mailbox_port: int,
                delegate_port: int,
                pub_key: str,
            ):
                return _make_libp2p_client_connection(
                    peer_public_key=pub_key, data_dir=".", node_port=delegate_port
                )

        elif connection == "mailbox":

            def con_maker(
                port: int,
                entry_peer: str,
                mailbox_port: int,
                delegate_port: int,
                pub_key: str,
            ):
                return _make_libp2p_mailbox_connection(
                    peer_public_key=pub_key, data_dir=".", node_port=mailbox_port
                )

        else:
            raise ValueError(f"Unsupported connection: {connection}")

        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            coro = _run(con_maker)
            first_time, second_time = asyncio.get_event_loop().run_until_complete(coro)

            return [
                ("first time (seconds)", first_time),
                ("second time (seconds)", second_time),
            ]
    finally:
        os.chdir(cwd)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
Bump the versions of AEA plugins throughout the code base.

    python scripts/update_plugin_versions.py --update "plugin-name,version" [--update ...]

Example of usage:

    python scripts/update_plugin_versions.py --update "open-aea-ledger-fetchai,0.2.0" --update "open-aea-ledger-ethereum,0.3.0"

"""

import argparse
import pprint
import re
import sys
from pathlib import Path
from typing import Dict, List, Tuple

from packaging.specifiers import SpecifierSet
from packaging.version import Version

from aea.cli.ipfs_hash import update_hashes
from aea.helpers.base import compute_specifier_from_version


ROOT_DIR = Path(__file__).parent.parent
PLUGINS_DIR = Path("plugins")
SETUP_PY_NAME_REGEX = re.compile(r"\Wname=\"(.*)\",")
SETUP_PY_VERSION_REGEX = re.compile(r"\Wversion=\"(.*)\",")


IGNORE_DIRS = [Path(".git")]


def update_plugin_setup(
    plugin_name: str, old_version: Version, new_version: Version
) -> bool:
    """Update plugin setup.py script with new version.

    :param plugin_name: the plugin name.
    :param old_version: the old version.
    :param new_version: the new version.
    :return: True if an update has been done, False otherwise.
    """
    setup_file = PLUGINS_DIR / plugin_name.strip("open-") / "setup.py"
    content = setup_file.read_text()
    new_content = re.sub(
        rf"version=['\"]{old_version}['\"],", f'version="{new_version}",', content
    )
    setup_file.write_text(new_content)
    return content != new_content


def process_plugin(
    plugin_name: str, old_version: Version, new_version: Version
) -> bool:
    """
    Process the plugin version.

    :param plugin_name: the plugin name.
    :param old_version: the old version.
    :param new_version: the new version.
    :return: True if an update has been done, False otherwise.
    """
    result = False
    result = update_plugin_setup(plugin_name, old_version, new_version) or result
    result = (
        update_plugin_version_specifiers(plugin_name, old_version, new_version)
        or result
    )
    return result


def update_plugin_version_specifiers(
    plugin_name: str, old_version: Version, new_version: Version
) -> bool:
    """
    Update aea_version specifier set in docs.

    :param plugin_name: the plugin name.
    :param old_version: the old version.
    :param new_version: the new version.
    :return: True if the update has been done, False otherwise.
    """
    old_specifier_set = compute_specifier_from_version(
        old_version, use_version_as_lower=True
    )
    new_specifier_set = compute_specifier_from_version(
        new_version, use_version_as_lower=True
    )
    print(f"Old version specifier: {old_specifier_set}")
    print(f"New version specifier: {new_specifier_set}")
    if old_specifier_set == new_specifier_set:
        print("Not updating version specifier - they haven't changed.")
        return False
    has_changed = False
    new_specifier_set = str(SpecifierSet(new_specifier_set))
    old_specifier_set = str(SpecifierSet(old_specifier_set))
    old_specifier_set_regex = re.compile(str(old_specifier_set).replace(" ", " *"))
    for file in filter(lambda p: not p.is_dir(), Path(".").rglob("*")):
        dir_root = Path(file.parts[0])
        if dir_root in IGNORE_DIRS:
            print(f"Skipping '{file}'...")
            continue
        print(
            f"Replacing '{old_specifier_set}' with '{new_specifier_set}' in '{file}'... ",
            end="",
        )
        try:
            content = file.read_text()
        except UnicodeDecodeError as e:
            print(f"Cannot read {file}: {str(e)}. Continue...")
        else:
            if old_specifier_set_regex.search(content) is None:
                print("No version to update found.")
                continue
            new_content = _replace_patterns(
                content, plugin_name, old_specifier_set, new_specifier_set
            )
            has_changed = has_changed or content != new_content
            file.write_text(new_content)
            print("Done!")
    return has_changed


def _replace_patterns(
    content: str, plugin_name: str, old_specifier: str, new_specifier: str
) -> str:
    """
    Replace specific patterns.

    It identifies three patterns:

    1) strings of the form:
    <plugin-name><old_specifier_set>
    2) YAML strings of the form:
    plugin-name:
      version: <old_specifier_set>
    3) strings of the form
    "<plugin-name>": {"version": "<old_specifier_set>"}

    :param content: the file content
    :param plugin_name: the plugin name
    :param old_specifier: the old specifier
    :param new_specifier: the new specifier
    :return: the new content.
    """
    # check pattern (1)
    content = re.sub(
        f"{plugin_name}{old_specifier}", f"{plugin_name}{new_specifier}", content
    )
    # check pattern (2)
    content = re.sub(
        f"({plugin_name}:\n *version: ){old_specifier}",
        rf"\g<1>{new_specifier}",
        content,
    )
    # check pattern (3)
    content = re.sub(
        f'"{plugin_name}": {{"version": "{old_specifier}"}}',
        f'"{plugin_name}": {{"version": "{new_specifier}"}}',
        content,
    )
    return content


def exit_with_message(message: str, exit_code: int = 1) -> None:
    """Exit the program with a message and an exit code."""
    print(message)
    sys.exit(exit_code)


def get_plugin_names_and_versions() -> Dict[str, Version]:
    """Get all the plugins names and versions."""
    result: Dict[str, Version] = {}
    for plugin_setup_script in PLUGINS_DIR.glob("*/setup.py"):
        content = plugin_setup_script.read_text()
        name_matches: List[str] = SETUP_PY_NAME_REGEX.findall(content)
        version_matches: List[str] = SETUP_PY_VERSION_REGEX.findall(content)
        if len(name_matches) != 1 or len(version_matches) != 1:
            exit_with_message(
                f"Unexpected result: in {result}, found plugin names: {name_matches} and versions: {version_matches}"
            )
        name, version = name_matches[0], version_matches[0]
        if name in result:
            print(f"Warning, duplicate plugin name: '{name}'.")
        result[name] = Version(version)
    return result


def name_version_pair(s: str) -> Tuple[str, str]:
    """
    Parse a name-version pair.

    :param s: the parameter string.
    :return: a pair of string (name, new_version)
    """
    try:
        name, version = [part.strip() for part in s.split(",")]
        return name, version
    except Exception:
        raise argparse.ArgumentTypeError(f"Name-version pair not correct: '{s}'")


def parse_args() -> argparse.Namespace:
    """Parse arguments."""
    parser = argparse.ArgumentParser("bump_aea_version")
    parser.add_argument(
        "--update",
        type=name_version_pair,
        metavar="'NAME,VERSION'",
        required=True,
        action="append",
        help="A comma-separated pair: 'plugin-name, new-version'.",
    )
    parser.add_argument("--no-fingerprint", action="store_true")
    arguments_ = parser.parse_args()
    return arguments_


def main() -> None:
    """Run the script."""
    arguments = parse_args()
    current_versions_by_name: Dict[str, Version] = get_plugin_names_and_versions()
    new_versions_by_name: Dict[str, Version] = dict(
        (name, Version(version)) for name, version in arguments.update
    )

    print(
        f"Found plugin names and versions:\n{pprint.pformat(current_versions_by_name)}"
    )
    print(f"Plugins to update:\n{pprint.pformat(new_versions_by_name)}")

    not_found_plugins = set(new_versions_by_name.keys()).difference(
        current_versions_by_name.keys()
    )
    if len(not_found_plugins) > 0:
        exit_with_message(
            f"Error: These plugins have not been found:\n{pprint.pformat(not_found_plugins)}"
        )

    have_updated_specifier_set = False

    for current_plugin_name, new_version in new_versions_by_name.items():
        old_version = current_versions_by_name[current_plugin_name]
        print(
            f"Processing {current_plugin_name}, old_version={old_version}, new_version={new_version}"
        )
        if new_version == old_version:
            print("Skipping, as old and new versions are equal.")
            continue

        have_updated_specifier_set = (
            process_plugin(current_plugin_name, old_version, new_version)
            or have_updated_specifier_set
        )

    return_code = 0
    if arguments.no_fingerprint:
        print("Not updating fingerprints, since --no-fingerprint was specified.")
    elif not have_updated_specifier_set:
        print("Not updating fingerprints, since no specifier set has been updated.")
    else:
        print("Updating hashes and fingerprints.")
        return_code = update_hashes(
            packages_dir=ROOT_DIR / "packages",
        )
    exit_with_message("Done!", exit_code=return_code)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains tests for P2PLibp2p connection."""

# pylint: skip-file

import itertools

import pytest

from aea.helpers.acn.uri import Uri

from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all
from packages.valory.connections.test_libp2p.tests.base import BaseP2PLibp2pTest


DEFAULT_NET_SIZE = 4


@pytest.mark.integration
@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionIntegrationTest(BaseP2PLibp2pTest):
    """Test mix of relay/delegate agents and client connections work together"""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        # relays
        main_relay = cls.make_connection()
        main_relay_peer = main_relay.node.multiaddrs[0]
        relay_2 = cls.make_connection(entry_peers=[main_relay_peer])
        relay_peer_2 = relay_2.node.multiaddrs[0]

        # delegates
        delegate_1 = cls.make_connection(
            entry_peers=[main_relay_peer], delegate=True, mailbox=True
        )
        delegate_2 = cls.make_connection(
            entry_peers=[relay_peer_2], delegate=True, mailbox=True
        )

        # agents
        cls.make_connection(entry_peers=[main_relay_peer], relay=False)
        cls.make_connection(entry_peers=[relay_peer_2], relay=False)

        # clients
        cls.make_client_connection(
            peer_public_key=delegate_1.node.pub,
            node_host=delegate_1.node.delegate_uri.host,
            node_port=delegate_1.node.delegate_uri.port,
        )
        cls.make_client_connection(
            peer_public_key=delegate_2.node.pub,
            node_host=delegate_2.node.delegate_uri.host,
            node_port=delegate_2.node.delegate_uri.port,
        )

        # mailboxes
        cls.make_mailbox_connection(
            peer_public_key=delegate_1.node.pub,
            node_host=Uri(delegate_1.node.mailbox_uri).host,
            node_port=Uri(delegate_1.node.mailbox_uri).port,
        )
        cls.make_mailbox_connection(
            peer_public_key=delegate_2.node.pub,
            node_host=Uri(delegate_2.node.mailbox_uri).host,
            node_port=Uri(delegate_2.node.mailbox_uri).port,
        )

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_send_and_receive(self) -> None:
        """Test envelope send/received by every pair of connection."""

        for sending, receiving in itertools.permutations(self.multiplexers, 2):
            sender = next(c.address for c in sending.connections)
            to = next(c.address for c in receiving.connections)
            envelope = self.enveloped_default_message(to=to, sender=sender)

            sending.put(envelope)
            delivered_envelope = receiving.get(block=True, timeout=10)
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)
